"""
update LW badpix table from C. Willott

- Add mask around edge of NIRCam detectors

https://github.com/gbrammer/grizli/issues/200
https://github.com/gbrammer/grizli/issues/233
"""

import os
import astropy.io.fits as pyfits
import numpy as np
from grizli import utils

bpix = {
    "along": np.zeros((2048, 2048), dtype=np.int16),
    "blong": np.zeros((2048, 2048), dtype=np.int16),
}

# Edge mask
nircam_edge = 8

for det in bpix:
    bpix[det][:nircam_edge, :] |= 1024
    bpix[det][-nircam_edge:, :] |= 1024
    bpix[det][:, :nircam_edge] |= 1024
    bpix[det][:, -nircam_edge:] |= 1024

willott_files = [
    "https://github.com/gbrammer/grizli/files/13878784/badpixlist_along_20240108.txt",
    "https://github.com/gbrammer/grizli/files/13878785/badpixlist_blong_20240108.txt",
    "https://github.com/user-attachments/files/15995936/badpixlist_along_20240626.txt",
    "https://github.com/user-attachments/files/15995938/badpixlist_blong_20240626.txt",
]

for file in willott_files:
    det = os.path.basename(file).split("_")[-2]
    new_data = utils.read_catalog(file, format="ascii")
    for x, y in zip(new_data["col1"], new_data["col2"]):
        bpix[det][y, x] |= 1

header = pyfits.Header()
header["INSTRUMENT"] = "NIRCAM"
header["TELESCOPE"] = "JWST"
header["NEDGE"] = (nircam_edge, "Number of masked pixels around edge")
header["EDGEVAL"] = (1024, "Edge DQ value")
header["BPVAL"] = (1, "Bad pixel DQ value")

for det in bpix:
    header["DETECTOR"] = ("NRC" + det).upper()
    out_file = f"nrc_badpix_240627_NRC{det.upper()}.fits.gz"
    print(f"NRC{det.upper()} N={(bpix[det] & 1 ).sum()}")
    pyfits.PrimaryHDU(header=header, data=bpix[det]).writeto(out_file, overwrite=True)

### Previous - add to earlier conservative bpix list
#
# old_list = ['nrc_badpix_231206_NRCALONG.fits.gz', 'nrc_badpix_231206_NRCBLONG.fits.gz']
# for new_file, old_file in zip(new_list, old_list):
#     det = os.path.basename(new_file).split('_')[1]
#
#     new_data = utils.read_catalog(new_file, format='ascii')
#
#     with pyfits.open(old_file) as im:
#         orig_bad = im[0].data.sum()
#
#         for x, y in zip(new_data['col1'], new_data['col2']):
#             im[0].data[y,x] |= 1
#
#         new_bad = im[0].data.sum()
#
#         msg = f'{det} prev = {orig_bad.sum()}  new = {new_bad.sum()}'
#         print(msg)
#
#     out_file = f'nrc_badpix_240112_NRC{det.upper()}.fits.gz'
#     im.writeto(out_file, overwrite=True)



from __future__ import division

import numpy as np
from numba import jit

DTYPE = float
ITYPE = int

__all__ = [
    "disperse_grism_object",
    "compute_segmentation_limits",
]


@jit(parallel=False, nopython=True, fastmath=True, error_model="numpy")
def disperse_grism_object(
    flam, segm, seg_id, idxl, yfrac, ysens, modelf, x0, shd, sh_thumb, shg
):
    """Compute a dispersed 2D spectrum

    Parameters
    ----------
    flam : array-like
        Direct image thumbnail

    segm : array-like
        Segmentation image

    seg_id : int, float
        Source ID.  The dispersed spectrum is computed for pixels in ``flam`` where
        ``segm == seg_id``.

    idxl : array-like (int)
        Flattened indices of the trace in the 2D cutout

    yfrac : array-like (float)
        Fraction of the flux of the "bottom" pixel along the beam

    ysens : array-like (float)
        Sensitivity or spectrum along the trace

    modelf : array-like (float)
        Flattened dispersed spectrum.  The dispersed spectrum is added in-place to this
        array

    x0 : int, int
        Reference pixel where the trace is defined

    shd : int, int
        Shape of the thumbnail array

    sh_thumb : int, int
        half-size of the thumbnail cutout to consider (this is generally ``x0``)

    shg : int, int
        Shape of the unflattened 2D spectrum

    Returns
    -------
    status : bool
        True if the function is executed successfully.  The dispersed spectrum itself
        is added to ``modelf``

    """
    nk = len(idxl)
    nl = len(modelf)

    ks = []
    for k in range(nk):
        if ysens[k] != 0:
            ks.append(k)

    for i in range(0 - sh_thumb[1], sh_thumb[1]):
        if (x0[1] + i < 0) | (x0[1] + i >= shd[1]):
            continue

        for j in range(0 - sh_thumb[0], sh_thumb[0]):
            if (x0[0] + j < 0) | (x0[0] + j >= shd[0]):
                continue

            fl_ij = flam[x0[0] + j, x0[1] + i]  # /1.e-17
            if fl_ij == 0:
                # Non-zero flux
                continue

            elif segm[x0[0] + j, x0[1] + i] != seg_id:
                # Segmentation map doesn't match the source id
                continue

            for k in ks:
                k0 = idxl[k]
                ysk = ysens[k]
                yfk = yfrac[k]

                k1 = k0 + j * shg[1] + i
                if (k1 >= 0) & (k1 < nl):
                    modelf[k1] += ysk * fl_ij * yfk

                k2 = k0 + (j - 1) * shg[1] + i
                if (k2 >= 0) & (k2 < nl):
                    modelf[k2] += ysk * fl_ij * (1 - yfk)

    return True


@jit(nopython=True, fastmath=True, error_model="numpy")
def compute_segmentation_limits(segm, seg_id, flam, shd):
    """Find pixel limits of a segmentation region

    Parameters
    ----------
    segm: ndarray (np.float32)
        segmentation array

    seg_id: int
        ID to test

    flam: ndarray (float)
        Flux array to compute weighted centroid within segmentation region

    shd: [int, int]
        Shape of segm

    Returns
    -------
    imin : int
        Minimium index of first array axis (y)

    imax : int
        Maximum index of first array axis (y)

    ic : float
        Weighted centroid along first array axis (y)

    jmin : int
        Minimium index of second array axis (x)

    jmax : int
        Maximum index of second array axis (x)

    jc : float
        Weighted centroid along second array axis (x)

    area : int
        Area of the segment

    flam_total : float
        Sum of ``flam`` within the segment

    """
    area = 0

    imin = shd[0]
    imax = 0
    jmin = shd[1]
    jmax = 0

    inumer = 0.0
    jnumer = 0.0
    flam_total = 0.0

    for i in range(shd[0]):
        for j in range(shd[1]):
            if segm[i, j] != seg_id:
                continue

            area += 1
            wht_ij = flam[i, j]
            inumer += i * wht_ij
            jnumer += j * wht_ij
            flam_total += wht_ij

            if i < imin:
                imin = i
            if i > imax:
                imax = i

            if j < jmin:
                jmin = j
            if j > jmax:
                jmax = j

    ### No matched pixels
    if flam_total == 0:
        flam_total = -99

    return (
        imin,
        imax,
        inumer / flam_total,
        jmin,
        jmax,
        jnumer / flam_total,
        area,
        flam_total,
    )


@jit(nopython=True, fastmath=True, error_model="numpy")
def seg_flux(flam, idxl, yfrac, ysens, full, x0, shd, shg):
    """
    Not used
    """
    pass



"""
Recalibrate JWST exposures with `snowblind` with the procedure as described there
"""
import os
import inspect
import warnings
import numpy as np

if os.path.exists('/GrizliImaging'):
    HOME = '/GrizliImaging'
else:
    HOME = os.getcwd()


def test():
    from grizli import prep
    
    file = 'jw01727039001_02101_00003_nrca2_rate.fits'
    file = 'jw02079004002_09201_00001_nrca2_rate.fits'
    file = 'jw01283004001_02201_00009_nrca1_rate.fits'
    
    do_recalibrate(rate_file='../RAW/' + file,
                   cores='all',
                   min_radius=3,
                   after_jumps=20,
                   context='jwst_1180.pmap',
                  )
    
    prep.fresh_flt_file(file, 
                        oneoverf_kwargs={'deg_pix':2048, 'dilate_iterations':3,
                                                               'thresholds':[5,4,3],
                                                               'other_axis':True})


def get_random_reprocess_file():
    """
    Get a table row of a file to be processed
    
    Returns
    -------
    res : None, table row
        Returns a single row from `reprocess_rates` with ``status == 0``, or None if 
        nothing found
    
    """
    from grizli.aws import db
    
    rows = db.SQL(f"""SELECT *
                           FROM reprocess_rates
                           WHERE status = 0 order by random() limit 1
                           """)
    
    if len(rows) == 0:
        return None
    else:
        return rows


def run_one(verbose=True, remove_result=True):
    """
    Run a single random file from `reprocess_rates` with ``status == 0``
    """
    import os
    import time
    import boto3
    
    from grizli.aws import db
    from grizli import utils
    
    row = get_random_reprocess_file()
    
    if row is None:
        with open(os.path.join(HOME, 'reprocess_finished.txt'),'w') as fp:
            fp.write(time.ctime() + '\n')
        
        return None
        
    print(f'============  Reprocess exposure  ==============')
    print(f"{row['rate_file'][0]}")
    print(f'========= {time.ctime()} ==========')
    
    with open(os.path.join(HOME, 'reprocess_history.txt'),'a') as fp:
        fp.write(f"{time.ctime()} {row['rate_file'][0]}\n")
    
    # Update db
    db.execute(f"""update reprocess_rates
               set status = 1, ctime={time.time()}
               where
               rate_file = '{row['rate_file'][0]}'
               AND prefix = '{row['prefix'][0]}'
               AND bucket = '{row['bucket'][0]}';
               """)
    
    #################
    # Run it
    #################
    try:
        status = do_recalibrate(**row[0])
    except:
        row['status'] = 5
        
        # Update db
        db.execute(f"""update reprocess_rates
                   set status = 3, ctime={time.time()}
                   where
                   rate_file = '{row['rate_file'][0]}'
                   AND prefix = '{row['prefix'][0]}'
                   AND bucket = '{row['bucket'][0]}';
                   """)
        
        return row
    
    ###############
    # Copy output
    ###############
    if status & os.path.exists(row['rate_file'][0]):
        
        s3 = boto3.resource('s3')
        bkt = s3.Bucket(row['bucket'][0])
        
        local_file = row['rate_file'][0]
        s3_prefix = os.path.join(row['prefix'][0], row['rate_file'][0])
        
        msg = f"Send {local_file} > s3://{row['bucket'][0]}/{s3_prefix}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
                    
        bkt.upload_file(local_file, s3_prefix, ExtraArgs={'ACL': row['acl'][0]},
                        Callback=None, Config=None)
                    
        row['status'] = 2
        row['ctime'] = time.time()
        
        # Update db
        db.execute(f"""update reprocess_rates
                   set status = 2, ctime={time.time()}
                   where
                   rate_file = '{row['rate_file'][0]}'
                   AND prefix = '{row['prefix'][0]}'
                   AND bucket = '{row['bucket'][0]}';
                   """)
        
        # Remove result
        if remove_result:
            msg = f"rm {row['rate_file'][0]}"
            utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
            
            os.remove(row['rate_file'][0])
            
    else:
        ###############
        # Problem
        ###############
        row['status'] = 3
        
        # Update db
        db.execute(f"""update reprocess_rates
                   set status = 3, ctime={time.time()}
                   where
                   rate_file = '{row['rate_file'][0]}'
                   AND prefix = '{row['prefix'][0]}'
                   AND bucket = '{row['bucket'][0]}';
                   """)
    return row


def do_recalibrate(rate_file='jw06541001001_03101_00001_nrs1_rate.fits', cores='half', min_radius=3, after_jumps=20, context=None, clean=True, skip_snowblind_after='1.14', **kwargs):
    """
    Recalibrate JWST uncal exposures with `snowblind` snowball masking
    
    Parameters
    ----------
    rate_file : str
        Filename of the output countrate exposure, must end in ``rate.fits``.
        The uncalibrated file will be fetched from MAST with 
        ``rate_file.replace('_rate.fits', '_uncal.fits')``
    
    cores : str
        Number of cores to use for the JWST pipeline steps, e.g., ``1``, ``quarter``,
        ``half``, ``all``
    
    min_radius, after_jumps : int, int
        `snowblind` parameters
    
    context : str
        Optional explicit CRDS context to use
    
    clean : bool
        Remove intermediate files
    
    Returns
    -------
    status : bool
        True if completed successfully.  False if ``rate_file`` doesn't have the 
        expected ``rate.fits`` suffix.  The ``rate_file`` product will be left in the 
        working directory.
    """
    from packaging.version import Version
    from snowblind import SnowblindStep
    import jwst
    from jwst.pipeline import Detector1Pipeline
    from jwst.step import RampFitStep
    from jwst.step import GainScaleStep
    
    import mastquery.utils

    try:
        from .. import jwst_utils
        from .. import utils
    except ImportError:
        from grizli import jwst_utils
        from grizli import utils
    
    frame = inspect.currentframe()
    _LOGFILE = utils.LOGFILE
    warnings.filterwarnings('ignore', category=RuntimeWarning)
    
    utils.LOGFILE = rate_file.split('.fits')[0] + '.log'
    utils.log_function_arguments(utils.LOGFILE, frame, 'do_recalibrate',
                                 ignore=['_LOGFILE'])
    
    if not rate_file.endswith('_rate.fits'):
        print(f'{rate_file} does not end with _rate.fits')
        return False
        
    if context is not None:
        jwst_utils.CRDS_CONTEXT = context
        jwst_utils.set_crds_context(verbose=False, override_environ=True)
    
    output_path, base_file = os.path.split(rate_file)
    mastquery.utils.download_from_mast([base_file.replace('_rate.fits','_uncal.fits')])
    
    # try:
    #     ncores = int(cores)
    # except ValueError:
    #     ncores = cores
    ncores = cores
    
    steps = {
        "jump": {
            "save_results": True,
            "maximum_cores": ncores
        },
        "ramp_fit": {
            "skip": True,
        },
        "gain_scale": {
            "skip": True,
        },
    }

    Detector1Pipeline.call(base_file.replace('_rate', '_uncal'), steps=steps)

    if Version(jwst.__version__) < Version(skip_snowblind_after):
        SnowblindStep.call(base_file.replace('_rate','_jump'),
                       min_radius=int(min_radius),
                       after_jumps=int(after_jumps),
                       save_results=True,
                       suffix="snowblind")
        next_exten = '_snowblind'
    else:
        msg = f'jwst {jwst.__version__} > {skip_snowblind_after}, skip snowblind'
        utils.log_comment(utils.LOGFILE, msg, verbose=True)
        
        next_exten = '_jump'
        
    rate, rateints = RampFitStep.call(base_file.replace('_rate', next_exten),
                                      maximum_cores=ncores)

    rate = GainScaleStep.call(rate)
    
    # Mask out nans and remove DQ=4
    bad = ~np.isfinite(rate.data)
    rate.dq[bad] |= 1
    rate.data[bad] = 0
    rate.dq -= (rate.dq & 4)

    rate.save(rate_file)
    
    if clean:
        for ext in ['_jump', '_snowblind', '_uncal', '_trapsfilled']:
            if os.path.exists(base_file.replace('_rate', ext)):
                print(f"Remove {base_file.replace('_rate', ext)}")
                os.remove(base_file.replace('_rate', ext))
            
    utils.LOGFILE = _LOGFILE
    
    return True



"""
Extract a NIRCam WFSS spectrum for a position on the sky
"""

import os
import numpy as np
import matplotlib.pyplot as plt

from .. import jwst_utils

jwst_utils.set_quiet_logging()

import astropy.io.fits as pyfits
import astropy.wcs as pywcs
import astropy.units as u
from astropy.coordinates import SkyCoord

import scipy.ndimage as nd

try:
    from skimage import morphology
except ImportError:
    morphology = None

from .. import utils, grismconf, multifit
from . import db, api_cutout

S3PATH = "s3://grizli-v2/HST/Pipeline/{assoc}/Prep/{dataset}_rate.fits"

THUMB = "https://grizli-cutout.herokuapp.com/thumb?all_filters=False&filters={filter}&ra={ra}&dec={dec}&size={size}&output=fits_weight"


def coordinate_string(ra=189.0706488, dec=62.2089502, precision=2):
    """
    Generate a sexagesimal string from decimal coordinates

    Parameters
    ----------
    ra, dec : float
        Decimal degrees R.A. and declination

    precision: int
        Precision factor for `~astropy.coordinates.SkyCoord.to_string`

    Returns
    -------
    center_coord : `~astropy.coordinates.SkyCoord`
        Coordinate object

    coostr : string
        Sexagesimal string

    """
    center_coord = SkyCoord(f"{ra} {dec}", unit="deg")

    coostr = center_coord.to_string(
        style="hmsdms",
        sep="",
        precision=precision,
    ).replace(" ", "")

    return center_coord, coostr


def thumbnail_cutout(
    ra=189.0706488,
    dec=62.2089502,
    cutout_size=(60 * 0.05 * u.arcsec),
    filter="F444W-CLEAR",
    local=True,
    prefix="cutout",
    skip_existing=True,
    verbose=True,
    **kwargs,
):
    """ """
    from astropy.io import fits
    from astropy.coordinates import SkyCoord

    center_coord, coostr = coordinate_string(ra=ra, dec=dec, precision=2)

    if "," in filter:
        fkey = "ir"
    else:
        fkey = filter.lower()

    cutout_file = f"{prefix}_{coostr}_{fkey}.fits"
    if os.path.exists(cutout_file) & skip_existing:
        msg = f"found thumbnail file: {cutout_file}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

        return cutout_file

    if hasattr(cutout_size, "unit"):
        size = cutout_size.to(u.arcsec).value
    else:
        size = cutout_size

    thumb_url = THUMB.format(ra=ra, dec=dec, size=size, filter=filter.lower())

    if local:
        kws = thumb_url.split("thumb?")[1].split("&")
        kwargs = {}
        for k in kws:
            key, value = k.split("=")
            kwargs[key] = value

        msg = f"api_cutout.get_thumb(**{kwargs}) -> {cutout_file}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

        hdu = api_cutout.get_thumb(**kwargs)
    else:
        msg = f"{thumb_url} -> {cutout_file}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
        hdu = fits.open(thumb_url)

    N = len(hdu) // 2
    if N > 2:
        num = hdu[0].data * 0.0
        den = hdu[0].data * 0
        for i in range(N):
            msg = f"thumbnail: {hdu[i*2+0].header['FILTER']}"
            utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
            num += hdu[i * 2 + 0].data * hdu[i * 2 + 1].data
            den += hdu[i * 2 + 1].data

        sci = num / den
        sci[den <= 0] = 0
        hdu[0].data = sci
        hdu[1].data = den

        fkey = "ir"
    else:
        fkey = hdu[0].header["FILTER"]

    cutout_file = f"{prefix}_{coostr}_{fkey}.fits"

    hdu[:2].writeto(cutout_file, overwrite=True)
    hdu.close()

    return cutout_file


def s3_cutout(
    ra=189.0706488,
    dec=62.2089502,
    cutout_size=(60 * 0.05 * u.arcsec),
    s3_file="s3://grizli-v2/Scratch/gdn-grizli-v7.4-f444w-clear_drc_sci.fits",
    prefix="cutout",
    ext=0,
    get_hdul=False,
    correct_wcs=False,
    verbose=False,
):
    """ """
    from astrocut import fits_cut
    from astropy.io import fits
    import astrocut.cutouts
    from astropy.coordinates import SkyCoord

    center_coord, coostr = coordinate_string(ra=ra, dec=dec, precision=2)

    fits_options = {}
    fits_options["use_fsspec"] = True
    fits_options["fsspec_kwargs"] = {"default_block_size": 10_000, "anon": True}

    cutout_file = f"{prefix}_{coostr}.fits"
    if os.path.exists(cutout_file) & (not get_hdul):
        return cutout_file

    msg = f"{s3_file} ({ra:.6f},{dec:.6f}) {cutout_size} -> {cutout_file}"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    with pyfits.open(s3_file, **fits_options) as hdul:
        cutout = astrocut.cutouts._hducut(
            hdul[ext],
            center_coord,
            (cutout_size, cutout_size),
            correct_wcs=correct_wcs,
            verbose=verbose,
        )

    cutout.header["COOSTR"] = (coostr, "Coordinate identifier")
    hdul = pyfits.PrimaryHDU(data=cutout.data, header=cutout.header)

    if get_hdul:
        return hdul

    hdul.writeto(cutout_file, overwrite=True)

    return cutout_file


def query_exposures(with_api=False, **kwargs):
    """
    Wrapper around query_exposures_api / db
    """
    if with_api:
        exp = query_exposures_api(**kwargs)
    else:
        exp = query_exposures_db(**kwargs)

    return exp


def query_exposures_api(
    ra=189.0706488,
    dec=62.2089502,
    radius=15.0,
    filters=["F356W-GRISMR", "F410M-GRISMR", "F444W-GRISMR"],
    verbose=False,
    **kwargs,
):
    """
    Query exposure metadata from API
    """
    if radius > 0:
        coo_str = f"polygon=rect({ra:.6f},{dec:.6f},{radius/60:.2f})"
    else:
        coo_str = f"coords={ra:.6f},{dec:.6f}"

    filt_str = ",".join(filters)

    API_URL = (
        "https://grizli-cutout.herokuapp.com/exposures?"
        + f"{coo_str}&filters={filt_str}&output=csv"
    )

    if verbose:
        print(API_URL)

    exp = utils.read_catalog(API_URL, format="csv")

    return exp


def query_exposures_db(
    ra=189.0706488,
    dec=62.2089502,
    radius=15.0,
    filters=["F356W-GRISMR", "F410M-GRISMR", "F444W-GRISMR"],
    verbose=False,
    **kwargs,
):
    """
    Query exposure metadata from DB
    """
    if radius > 0:
        exposure_query = f"""
        select * from exposure_files 
        where polygon(footprint) && polygon(circle(point({ra},{dec}), {radius/3600}))
        AND filter in ({(','.join(db.quoted_strings(filters))).upper()})
        order by assoc, dataset
        """

    else:
        exposure_query = f"""
    select * from exposure_files 
    where polygon(footprint) @> point({ra},{dec})
    AND filter in ({(','.join(db.quoted_strings(filters))).upper()})
    order by assoc, dataset
    """

    if verbose:
        print(exposure_query)

    exp = db.SQL(exposure_query)

    return exp


def id_from_segmentation(segmentation_image, ra, dec, verbose=True):
    """
    Get ID from segmentation image

    Parameters
    ----------
    segmentation_image : str
        Filename of segmentation image

    ra, dec : float
        Coordinates in decimal degrees

    Returns
    -------
    segment_id : int
        Value of `segmentation_image` at specified coordinates

    """
    with pyfits.open(segmentation_image) as im:
        wcs = pywcs.WCS(im[0].header)
        xp, yp = np.squeeze(wcs.all_world2pix([ra], [dec], 0)).astype(int)
        try:
            segment_id = im[0].data[yp, xp]
            msg = f"{segmentation_image} @ ({ra:.5f},{dec:.5f}) = {segment_id}"
        except IndexError:
            segment_id = 0
            msg = f"({ra:.5f},{dec:.5f}) not in {segmentation_image}"

    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    return segment_id


def simple_zeroth_mask(grp, dilation=None, erosion=8, update=True, verbose=True, **kwargs):
    """
    Compute a simplified mask for zeroth orders shifting the segmentation image

    Parameters
    ----------
    grp : `~grizli.multifit.GroupFLT`
        Exposure group object

    dilation : int, None
        Number of mask dilation iterations

    erosion : int, None
        Number of mask erosion iterations

    update : bool
        Apply mask to DQ, SCI, ERR arrays

    Returns
    -------
    mask : array-like
        Boolean zeroth order mask calculated by shifting the non-zero segmentation mask
        by the expected zeroth order location as defined in the center of the detector

    """
    for flt in grp.FLTs:
        if flt.seg.max() == 0:
            msg = "simple_zeroth_mask: no non-zero segmentation pixels"
            utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
            return None
        elif "B" not in flt.conf.beams:
            msg = "zeroth order (B) not found in config"
            utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
            return None

        dx = flt.conf.dxlam["B"]
        dy, lam = flt.conf.get_beam_trace(1024, 1024, dx=dx, beam="B")
        N = len(dx)
        dxp = int(dx[N // 2])
        dyp = int(dy[N // 2])

        msg = f"simple_zeroth_mask: dx={dxp} dy={dyp}   erosion={erosion} dilation={dilation}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

        mask = flt.seg > 0
        if dilation is not None:
            if morphology is None:
                mask = nd.binary_dilation(mask, iterations=dilation)
            else:
                mask = morphology.isotropic_dilation(mask, dilation)
        if erosion is not None:
            if morphology is None:
                mask = nd.binary_erosion(mask, iterations=erosion)
            else:
                mask = morphology.isotropic_erosion(mask, erosion)

        mask = np.roll(np.roll(mask, dxp, axis=1), dyp, axis=0)

        if update:
            flt.grism.data["DQ"][mask] |= 4096
            flt.grism.data["ERR"][mask] *= 0
            flt.grism.data["SCI"][mask] *= 0

    return mask


def extract_from_coords(
    ra=189.0706488,
    dec=62.2089502,
    grisms=["F356W-GRISMR", "F410M-GRISMR", "F444W-GRISMR"],
    query_with_api=True,
    size=48,
    grp=None,
    clean=False,
    get_cutout=True,
    cutout_filter="F444W-CLEAR",
    segmentation_image=None,
    source_id=0,
    prefix="gdn-grism",
    verbose=True,
    mb_kwargs={},
    thumbnail_size=None,
    filter_kwargs=None,
    savefig=True,
    savefits=True,
    zeroth_mask_kwargs={"erosion": 8},
    use_jwst_crds=False,
    **kwargs,
):
    """
    Query and download slitless exposures
    """

    center_coord, coostr = coordinate_string(ra=ra, dec=dec, precision=2)

    group_name = f"{prefix}_{coostr}"

    exp = query_exposures(
        ra=ra,
        dec=dec,
        filters=grisms,
        with_api=query_with_api,
        verbose=verbose,
        **kwargs,
    )

    rate_files = []

    utils.LOGFILE = f"{group_name}.log.txt"
    msg = f"extract_from_coords: {group_name} {len(exp)} exposures"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    for row in exp:
        rate_file = db.download_s3_file(
            S3PATH.format(**row), overwrite=False, verbose=verbose, **kwargs
        )
        if os.path.exists(rate_file):
            rate_files.append(rate_file)

    if get_cutout:
        if thumbnail_size == None:
            cutout_size = size * 0.04 * u.arcsec
        else:
            cutout_size = thumbnail_size * u.arcsec

        cutout_file = thumbnail_cutout(
            ra=ra,
            dec=dec,
            cutout_size=cutout_size,
            verbose=True,
            prefix=prefix,
            filter=cutout_filter,
            **kwargs,
        )
    else:
        cutout_file = None

    msg = f"extract_from_coords: direct image = {cutout_file}"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    if (segmentation_image is not None) & (source_id == 0):
        # Get source from segmentation ID
        source_id = id_from_segmentation(segmentation_image, ra, dec, verbose=verbose)

    if grp is None:
        beams = []
        for j, file in enumerate(rate_files):
            if (verbose & 1) > 0:
                msg = f"{j+1:>2} / {len(rate_files)} GroupFLT {os.path.basename(file)}"
                utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

            grp = multifit.GroupFLT(
                grism_files=[file],
                pad=[800, 800],
                ref_file=cutout_file,
                verbose=((verbose & 2) > 0),
                cpu_count=-1,
                seg_file=segmentation_image,
                use_jwst_crds=use_jwst_crds,
            )

            if (zeroth_mask_kwargs is not None) & (segmentation_image is not None):
                # print('x before mask', (grp.FLTs[0].grism['DQ'] > 0).sum())
                _ = simple_zeroth_mask(grp, **zeroth_mask_kwargs)
                # print('x  after mask', (grp.FLTs[0].grism['DQ'] > 0).sum())

            if filter_kwargs is not None:
                grp.subtract_median_filter(**filter_kwargs)

            beams += grp.get_beams(
                source_id, center_rd=(ra, dec), size=size, **mb_kwargs
            )

            del(grp)

    else:
        beams = grp.get_beams(source_id, center_rd=(ra, dec), size=size, **mb_kwargs)

    msg = f"extract_from_coords: {group_name} {len(beams)} beam cutouts"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    mb = multifit.MultiBeam(
        beams,
        group_name=group_name,
        **mb_kwargs,
    )

    # for b in mb.beams:
    #     msk = b.grism['ERR'] > 0
    #     b.grism.data['SCI'] -= np.nanpercentile(b.grism['SCI'][msk], 40)

    _hdu, _fig2d = mb.drizzle_grisms_and_PAs(
        fig_args={
            "mask_segmentation": True,
            "average_only": False,
            "scale_size": 1,
            "cmap": "bone_r",
        }
    )

    _fig1d = mb.oned_figure(
        figsize=(10, 4), units="ujy", bin=2, ylim_percentile=0.2, show_beams=True
    )

    if savefig:
        _fig2d.savefig(f"{group_name}.2d.png")
        _fig1d.savefig(f"{group_name}.1d.png")

    if savefits:
        mb_hdul = mb.write_master_fits(get_trace_table=True, get_hdu=True)
        mb_hdul.writeto(f"{group_name}.beams.fits", overwrite=True)

        oned = mb.oned_spectrum()
        oned_hdu = mb.oned_spectrum_to_hdu()
        oned_hdu.writeto(f"{group_name}.1d.fits", overwrite=True)

    if clean:
        for file in rate_files:
            if os.path.exists(file):
                os.remove(file)

    return mb


def combine_beams_2d(
    mb,
    step=0.5,
    pixfrac=0.75,
    ymax=12.5,
    bkg_percentile=None,
    profile_type="grizli",
    profile_sigma=1.0,
    profile_offset=0.0,
    cont_spline=41,
    zfit_nspline=-1,
    z=None,
    savefig=True,
    verbose=True,
    sys_err=0.03,
    auto_niriss=False,
    zfit_kwargs={},
    ylim=(-3, 10),
    yticks=[0, 3, 5, 7],
    **kwargs,
):
    """
    Create rectified 2D spectra with `msaexp.slit_combine.pseudo_drizzle`

    Parameters
    ----------
    mb : `grizli.multifit.MultiBeam`
        Grizli "beams" object

    step : float
        Step size of spatial axis relative to the native pixels

    pixfrac : float
        Pseudo-drizzle pixfrac value

    ymax : float
        Extent of the cross-dispersion size to plot

    bkg_percentile : float, None
        Percentile of the 2D pixel distribution to use as a pedestal background

    profile_type : ("grizli", "gaussian")
        Extraction profile to use.  With ``grizli``, use the dispersed 2D model
        derived from the direct image cutout.  With ``gaussian``, use a pixel-integrated
        Gaussian with the ``profile_sigma`` and ``profile_offset`` parameters.

    profile_sigma, profile_offset : float, float
        Gaussian profile width and center relative to the expected trace,
        in units of the native pixels

    cont_spline : int
        Degree of the cubic spline used to model the continuum / contamination

    z : float
        Trial redshift used for emission line labels

    Returns
    -------
    b2d : dict
        Derived products with keys of the available grism names.

    """
    import msaexp.utils
    import msaexp.slit_combine
    import msaexp.resample_numba

    mb.compute_model()

    if mb.beams[0].grism.filter.startswith("GR") & auto_niriss:
        if zfit_kwargs is not None:
            if "velocity_sigma" not in zfit_kwargs:
                dv = 800
            else:
                dv = zfit_kwargs["velocity_sigma"]

            zfit_kwargs["velocity_sigma"] = dv
            zfit_kwargs["dz"] = dv / 3.0e5 / 2

        cont_spline = 0
        zfit_nspline = 11

    b2d = {}
    for gr in mb.PA:
        b2d[gr] = {
            "group_name": mb.group_name,
            "sci": [],
            "wht": [],
            "prof": [],
            "bkg": [],
            "idx": [],
            "s1d": [],
            "e1d": [],
            "c1d": [],
            "contam": [],
            "cont_model": [],
        }

        for j, pa in enumerate(mb.PA[gr]):
            beams = [mb.beams[i] for i in mb.PA[gr][pa]]
            b2d[gr]["idx"] += mb.PA[gr][pa]

            ydata = []
            ldata = []
            sdata = []
            vdata = []
            mdata = []

            for beam in beams:
                size = mb.beams[0].sh[0] // 2
                yp, xp = np.indices(beam.sh) * 1.0
                yp += 0.5 - size - beam.beam.ytrace

                lp = xp * 0 + beam.beam.lam / 1.0e4
                ok = np.abs(yp) < ymax
                oktr = np.abs(yp) > 3
                ok &= (beam.grism["DQ"] & (1 + 1024 + 4096)) == 0
                ok &= beam.grism["SCI"] != 0
                ok &= beam.grism["ERR"] > 0
                sok = beam.beam.sensitivity > 0.01 * beam.beam.sensitivity.max()
                ok &= sok

                if bkg_percentile is not None:
                    bkg = np.nanpercentile(beam.grism["SCI"][ok & oktr], bkg_percentile)
                else:
                    bkg = 0.0

                if (bkg > 5 * np.median(beam.grism["ERR"][ok])) & (1):
                    bkg = 0.0

                b2d[gr]["bkg"].append(bkg)

                ydata.append(1 * yp[ok])  # Not sure about the sign
                ldata.append(lp[ok])
                sdata.append(beam.grism["SCI"][ok] - bkg)
                vdata.append(beam.grism["ERR"][ok] ** 2)
                mdata.append(beam.flat_flam.reshape(beam.sh)[ok])

            yb = np.arange(-ymax, ymax + 0.1, step)
            wgrid = np.arange(
                *utils.GRISM_LIMITS[gr][:2], utils.GRISM_LIMITS[gr][2] / 1.0e4
            )
            wb = msaexp.utils.array_to_bin_edges(wgrid)

            b2d[gr]["yb"] = yb
            b2d[gr]["wgrid"] = wgrid
            b2d[gr]["wb"] = wb

            ldata = np.hstack(ldata)
            ydata = np.hstack(ydata)
            sdata = np.hstack(sdata)
            vdata = np.hstack(vdata)
            wdata = 1.0 / vdata
            mdata = np.hstack(mdata)

            if 0:
                msk = np.abs(ydata) > 2
                msk *= (sdata - np.nanmedian(sdata)) < 2 * np.sqrt(vdata)
                cbkg = np.polyfit(ldata[msk], sdata[msk], 1)
                sdata -= np.polyval(cbkg, ldata)

            drz = msaexp.slit_combine.pseudo_drizzle(
                ldata,
                ydata,
                sdata,
                vdata,
                wdata,
                wb,
                yb,
                pixfrac=pixfrac,
            )

            if profile_type == "grizli":
                prof = mdata
                to_ujy = (mdata**0 * u.erg / u.second / u.cm**2 / u.Angstrom).to(
                    u.microJansky, equivalencies=u.spectral_density(ldata * u.micron)
                )
                prof /= to_ujy.value
                if 0:
                    prof /= np.interp(
                        np.hstack(ldata) * 1.0e4, beam.beam.lam, beam.beam.sensitivity
                    )
                prof[~np.isfinite(prof)] = 0.0

            elif profile_type == "gaussian":
                prof = msaexp.resample_numba.pixel_integrated_gaussian_numba(
                    ydata, profile_offset, profile_sigma, dx=1.0
                )

            else:
                b = beams[0]
                # plt.imshow(np.log(b.direct['REF']))

                yprof = np.nansum(b.direct["REF"], axis=1)
                yprof /= yprof.sum()

                import msaexp.resample_numba

                xst = 1.0
                # size = 48
                N = 10

                xarr = np.arange(len(yprof)) * 1.0 - size
                steps = np.linspace(-xst * N, xst * N, N * 2 + 1)

                comps = np.array(
                    [
                        msaexp.resample_numba.pixel_integrated_gaussian_numba(
                            xarr, st, xst, dx=1.0
                        )
                        for st in steps
                    ]
                ).T

                c = np.linalg.lstsq(comps, yprof, rcond=None)

                cprof = np.array(
                    [
                        msaexp.resample_numba.pixel_integrated_gaussian_numba(
                            ydata, st + profile_offset, xst, dx=1.0
                        )
                        for st in steps
                    ]
                ).T
                prof = cprof.dot(c[0] / c[0].sum())

            pdrz = msaexp.slit_combine.pseudo_drizzle(
                ldata,
                ydata,
                prof,
                vdata,
                wdata,
                wb,
                yb,
                pixfrac=0.75,
            )
            p2d = pdrz[0] / pdrz[2]

            sci = drz[0] / drz[2]
            b2d[gr]["sci"].append(sci)
            b2d[gr]["wht"].append(drz[2])
            b2d[gr]["prof"].append(p2d)

    for gr in mb.PA:

        NPA = len(mb.PA[gr])
        fig, axes = plt.subplots(
            1 + NPA * 2,
            1,
            figsize=(8, 1.5 + 2 * NPA),
            sharex=True,
            height_ratios=[1, 1] * NPA + [2],
        )

        # ok = (b2d[gr]["wht"] > 0) & np.isfinite(b2d[gr]["sci"])
        vmax = np.nanpercentile(np.array(b2d[gr]["sci"]), 80)
        avg_wht = np.nanmedian(np.array(b2d[gr]["wht"]))
        vmax = np.maximum(vmax, 3 / np.sqrt(avg_wht))

        yb = b2d[gr]["yb"]
        wgrid = b2d[gr]["wgrid"]
        wb = b2d[gr]["wb"]

        smax = 0.0
        b2d[gr]["wminmax"] = (wgrid.min(), wgrid.max())

        for j, pa in enumerate(mb.PA[gr]):
            sci = b2d[gr]["sci"][j]
            p2d = b2d[gr]["prof"][j]
            wht = b2d[gr]["wht"][j]

            if cont_spline > 0:
                w2 = (wgrid[:, None] * np.ones(sci.shape[0])).T

                okw = wht > 0
                wminmax = (w2[okw].min(), w2[okw].max())
                wminmax = b2d[gr]["wminmax"]

                bspl = utils.bspline_templates(
                    w2.flatten(), df=cont_spline, minmax=wminmax, get_matrix=True
                )
                A = np.vstack(
                    [
                        # np.polynomial.polynomial.polyvander(w2.flatten(), 0).T,
                        p2d.flatten()
                        * bspl.T
                    ]
                )

                test_resid = True
                cont_model = 0.0

                for _iter in range(2):
                    okw = (wht > 0) & test_resid
                    okw &= sci * np.sqrt(wht) > -4
                    bspl = utils.bspline_templates(
                        w2[okw], df=cont_spline, minmax=wminmax, get_matrix=True
                    )
                    Ax = np.vstack(
                        [
                            # np.polynomial.polynomial.polyvander(w2[okw], 0).T,
                            (p2d * np.sqrt(wht))[okw]
                            * bspl.T
                        ]
                    )
                    y = sci[okw] * np.sqrt(wht[okw])
                    cspl = np.linalg.lstsq(Ax.T, y, rcond=None)
                    cont_model = A.T.dot(cspl[0]).reshape(w2.shape)
                    test_resid = (sci - cont_model) * np.sqrt(wht) < 2
                    # test_resid &= nd.binary_erosion(test_resid, iterations=2)

                    # bad_resid = (sci - cont_model) * np.sqrt(wht) > 4
                    # bad_resid &= nd.binary_erosion(bad_resid, iterations=1)
                    # test_resid = ~nd.binary_dilation(bad_resid, iterations=2)

            else:
                cont_model = 0.0

            opt_num = np.nansum((sci - cont_model) * p2d * wht, axis=0)
            opt_den = np.nansum(p2d**2 * wht, axis=0)

            s1d = opt_num / opt_den
            e1d = np.sqrt(1.0 / opt_den)

            if cont_spline > 0:
                A = utils.bspline_templates(
                    wgrid,
                    minmax=b2d[gr]["wminmax"],  # (wgrid.min(), wgrid.max()),
                    df=cont_spline,
                    get_matrix=True,
                ).T

                test_resid = True
                c1d = 0.0

                for _iter in range(2):
                    okw = (opt_den > 0) & test_resid
                    okw &= s1d * np.sqrt(opt_den) > -4
                    bspl = utils.bspline_templates(
                        wgrid[okw],
                        minmax=b2d[gr]["wminmax"],  # (wgrid.min(), wgrid.max()),
                        df=cont_spline,
                        get_matrix=True,
                    )
                    Ax = np.vstack(
                        [
                            # np.polynomial.polynomial.polyvander(w2[okw], 0).T,
                            np.sqrt(opt_den)[okw]
                            * bspl.T
                        ]
                    )
                    y = s1d[okw] * np.sqrt(opt_den[okw])
                    cspl = np.linalg.lstsq(Ax.T, y, rcond=None)
                    c1d = A.T.dot(cspl[0])
                    test_resid = (s1d - c1d) * np.sqrt(opt_den) < 2

                # c1d = cont_model
                s1d -= c1d
            else:
                c1d = np.nansum((sci * 0.0 + cont_model) * p2d * wht, axis=0) / opt_den

            contam = np.nansum((sci * 0.0 + cont_model) * p2d * wht, axis=0) / opt_den

            total_err = np.sqrt(e1d**2 + (sys_err * contam) ** 2)
            # total_err = e1d

            pl = axes[-1].plot(wgrid, s1d / total_err, alpha=0.4, zorder=10)

            axes[-1].plot(
                wgrid, c1d / total_err, alpha=0.2, color=pl[0].get_color(), zorder=1
            )

            axes[-1].fill_between(
                wgrid, e1d**0, -(e1d**0), alpha=0.08, color=pl[0].get_color()
            )

            b2d[gr]["s1d"].append(s1d)
            b2d[gr]["e1d"].append(total_err)
            b2d[gr]["c1d"].append(c1d)
            b2d[gr]["contam"].append(contam)
            b2d[gr]["cont_model"].append(cont_model)

            nmad = 1.48 * np.nanmedian(np.abs(s1d - np.nanmedian(s1d)))
            # smax = np.max([smax, np.nanpercentile(s1d, 80), 3*nmad])
            smax = np.max([smax, np.nanmedian(e1d[(e1d > 0) & (np.isfinite(e1d))])])
            # print(smax)
            # smax = np.maximum(smax, 3*nmad)

            ax = axes[j * 2]
            ax.imshow(
                sci - cont_model * 0,
                aspect="auto",
                extent=(wb[0], wb[-1], yb[0], yb[-1]),
                vmin=-0.1 * vmax,
                vmax=2 * vmax,
                cmap="bone_r",
            )
            ax.grid()
            ax.text(
                0.98,
                0.9,
                f"PA: {pa:.0f}",
                ha="right",
                va="top",
                fontsize=8,
                transform=ax.transAxes,
                bbox={"ec": "None", "fc": "w", "alpha": 1.0},
            )

            m2d = s1d * p2d  # + cont_model
            ax = axes[j * 2 + 1]
            ax.imshow(
                m2d,
                aspect="auto",
                extent=(wb[0], wb[-1], yb[0], yb[-1]),
                vmin=-0.1 * vmax,
                vmax=2 * vmax,
                cmap="bone_r",
            )
            ax.grid()
            for k in [0, 1]:
                axes[j * 2 + k].set_ylim(-ymax, ymax + 1)
                axes[j * 2 + k].set_yticklabels([])

        scl = 2
        axes[-1].set_ylim(*ylim)
        axes[-1].set_yticks(yticks)
        axes[-1].set_ylabel(r"$\sigma$")
        axes[-1].grid()
        axes[-1].set_xlabel(r"$\lambda$ - " + f"{gr}")

        data = b2d[gr]
        sp = utils.GTable()
        sp["wave"] = data["wgrid"] * u.micron
        wht = 1.0 / np.array(data["e1d"]) ** 2
        sp["flux"] = (
            np.nansum(np.array(data["s1d"]) * wht, axis=0) / np.nansum(wht, axis=0)
        ) * u.microJansky

        sp["contam"] = (
            np.nansum(np.array(data["contam"]) * wht, axis=0) / np.nansum(wht, axis=0)
        ) * u.microJansky

        sp["err"] = 1.0 / np.sqrt(np.nansum(wht, axis=0)) * u.microJansky

        sp["bkg"] = (
            np.nansum(np.array(data["c1d"]) * wht, axis=0) / np.nansum(wht, axis=0)
        ) * u.microJansky

        sp.meta["GRATING"] = "GRISMR"
        sp.meta["FILTER"] = gr
        sp.meta["INSTRUME"] = "NIRCAM"
        sp.meta["RA"] = mb.ra
        sp.meta["DEC"] = mb.dec
        sp.meta["SPLWMIN"] = b2d[gr]["wminmax"][0]
        sp.meta["SPLWMAX"] = b2d[gr]["wminmax"][1]

        spec_file = f"{mb.group_name}.{gr}.spec.fits".lower()
        msg = f"final spectrum file: {spec_file}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

        sp.write(spec_file, overwrite=True)
        b2d[gr]["spec"] = sp
        b2d[gr]["spec_file"] = spec_file

        axes[-1].plot(wgrid, sp["flux"] / sp["err"], alpha=0.3, color="k", zorder=100)

        b2d[gr]["fig"] = fig

    if zfit_kwargs is not None:
        if zfit_nspline < 0:
            zfit_nspline = cont_spline

        zres = redshift_fit_1d(b2d, nspline=zfit_nspline, **zfit_kwargs)
        z = zres["z"]
    else:
        zres = None

    if z is not None:
        for gr in b2d:
            axes = b2d[gr]["fig"].axes

            for ax in axes:
                yl = ax.get_ylim()
                xl = ax.get_xlim()
                lw, lr = utils.get_line_wavelengths()
                lines = [
                    "OIII-4363",
                    "NeIII-3968",
                    "NeIII-3867",
                    "OII",
                    "OIII",
                    "Hb",
                    "Hg",
                    "Hd",
                    "Ha+NII",
                    "SII",
                    "SIII",
                    "PaB",
                    "PaG",
                    "PaD",
                    "HeI-1083",
                ]
                rest_wave = []
                for li in lines:
                    rest_wave += lw[li]

                ax.vlines(
                    np.array(rest_wave) * (1 + z) / 1.0e4,
                    *yl,
                    color="magenta",
                    linestyle=":",
                    alpha=0.3,
                )
                ax.set_ylim(*yl)
                ax.set_xlim(*xl)

            ax.text(
                0.98,
                0.9,
                f"z = {z:.4f}",
                ha="right",
                va="top",
                fontsize=8,
                color="magenta",
                transform=ax.transAxes,
                bbox={"ec": "None", "fc": "w", "alpha": 1.0},
            )

    for gr in b2d:
        fig = b2d[gr]["fig"]
        fig.tight_layout(pad=0.5)
        if mb.id > 0:
            label = f"{mb.id}\n{mb.group_name}"
        else:
            label = mb.group_name

        utils.figure_timestamp(fig, text_prefix=label + "\n")

        if savefig:
            fig.savefig(f"{mb.group_name}.{gr.lower()}.png")

    return b2d, zres


def redshift_fit_1d(
    b2d,
    zgrid=None,
    dz=0.0003,
    rest_wave=[1.3e4, 4800],
    scale_disp=0.5,
    verbose=True,
    velocity_sigma=None,
    savefig=True,
    nspline=0,
    spline_type="add",
    **kwargs,
):
    """
    Fit redshift with 1D spectra
    """
    from scipy.stats import norm
    from msaexp.spectrum import SpectrumSampler
    import eazy.templates

    templ = eazy.templates.Template("fsps_line_templ.fits")
    null_template = eazy.templates.Template(
        arrays=(templ.wave, np.ones_like(templ.flux_flam()))
    )

    loss_null = 0.0
    specs = {}
    loss = {}
    splines = {}

    for gi in b2d:
        spec = SpectrumSampler(b2d[gi]["spec_file"])
        group_name = b2d[gi]["group_name"]

        loss[gi] = norm(loc=0, scale=spec["full_err"][spec.valid])
        if nspline <= 0:
            splines[gi] = None
            null_i = loss[gi].logpdf(spec["flux"][spec.valid]).sum()
        else:
            splines[gi] = spec.bspline_array(
                nspline=nspline,
                minmax=(spec.meta["SPLWMIN"] * 1.0e4, spec.meta["SPLWMAX"] * 1.0e4),
                by_wavelength=True,
                orders=[1],
            )

            tfit = spec.fit_single_template(
                null_template,
                z=0,
                scale_disp=scale_disp,
                velocity_sigma=velocity_sigma,
                spl=splines[gi],
                spline_type=spline_type,
            )
            null_i = loss[gi].logpdf((spec["flux"] - tfit["model"])[spec.valid]).sum()
            # null_i = loss[gi].logpdf(spec["flux"][spec.valid]).sum()

        spec.spec.meta["LOSSNULL"] = (null_i, "Null lnP")
        loss_null += null_i

        specs[gi] = spec

    if zgrid is None:
        zlim = [1000, 0]
        for gi in specs:
            glim = utils.GRISM_LIMITS[gi]
            zgr = [glim[i] * 1.0e4 / rest_wave[i] - 1 for i in [0, 1]]

            zlim[0] = np.minimum(zlim[0], zgr[0])
            zlim[1] = np.maximum(zlim[1], zgr[1])

        zgrid = utils.log_zgrid(zlim, dz)

    msg = f"redshift_fit_1d {group_name} z=[{zgrid.min():.3f}, {zgrid.max():.3f}]"
    msg += f" dz={dz}  nsteps={len(zgrid)}"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    lnp = zgrid * 0.0

    if velocity_sigma is None:
        velocity_sigma = dz * 3.0e5

    for j, zi in enumerate(zgrid):
        for gi in specs:
            tfit = specs[gi].fit_single_template(
                templ,
                z=zi,
                scale_disp=scale_disp,
                velocity_sigma=velocity_sigma,
                spl=splines[gi],
                spline_type=spline_type,
            )
            lnp[j] += tfit["lnp"].sum()

    fig, ax = plt.subplots(1, 1, figsize=(8, 3.5))
    ax.plot(zgrid, lnp - loss_null, color="0.4", alpha=0.5)
    ax.grid()

    iz = np.argmax(lnp)
    zbest = zgrid[iz]

    dlnp = np.nanmax(lnp) - loss_null

    msg = f"redshift_fit_1d {group_name} best z={zbest:.5f}  dlnP={dlnp:.1f}"
    utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

    loss_z = 0.0
    for gi in specs:
        spec = specs[gi]

        tfit = spec.fit_single_template(
            templ,
            z=zbest,
            scale_disp=scale_disp,
            velocity_sigma=velocity_sigma,
            spl=splines[gi],
            spline_type=spline_type,
        )

        b2d[gi]["fig"].axes[-1].plot(
            spec["wave"], tfit["model"] / spec["full_err"], color="magenta", alpha=0.5
        )
        spec.spec["model"] = tfit["model"]
        spec.spec.meta["REDSHIFT"] = (zbest, "Best redshift")

        if (splines[gi] is not None) & (spline_type == "add"):
            spline_continuum = tfit["A"].T[:, 1:].dot(tfit["coeffs"][1:])
            spec.spec["continuum"] = spline_continuum
            b2d[gi]["fig"].axes[-1].plot(
                spec["wave"],
                (spec["flux"] - spline_continuum) / spec["full_err"],
                color="0.5",
                alpha=0.5,
            )

            b2d[gi]["fig"].axes[-1].plot(
                spec["wave"],
                (spec["model"] - spline_continuum) / spec["full_err"],
                color="tomato",
                alpha=0.5,
            )

        loss_i = loss[gi].logpdf((spec["flux"] - tfit["model"])[spec.valid]).sum()

        dlnp_i = loss_i - spec.spec.meta["LOSSNULL"][0]
        spec.spec.meta["DLNP"] = dlnp_i

        spec.spec.meta["LOSSZ"] = (loss_i, "lnP at best redshift")
        loss_z += loss_i

        msg = f"redshift_fit_1d {group_name} {gi} dlnP={dlnp_i:.1f}"
        utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

        spec.spec.write(b2d[gi]["spec_file"], overwrite=True)

    yl = ax.get_ylim()

    ax.text(
        zbest,
        lnp[iz] - loss_null + 0.02 * (yl[1] - yl[0]),
        f"{zbest:.4f}",
        color="tomato",
        ha="center",
        va="bottom",
        fontsize=8,
    )

    ax.set_ylim(yl[0], yl[0] + 1.05 * (yl[1] - yl[0]))

    ax.set_xlabel("z")
    ax.set_ylabel(r"$\Delta~\log{P(z)}$")
    xl = ax.get_xlim()
    ax.hlines([loss_null * 0], *xl, color="pink", alpha=0.5)
    ax.set_xlim(*xl)

    fig.tight_layout(pad=0.5)

    utils.figure_timestamp(
        fig,
        text_prefix=(group_name + "\n"),
    )

    res = {
        "zgrid": zgrid,
        "lnp": lnp,
        "loss_null": loss_null,
        "z": zbest,
        "fig": fig,
    }

    if savefig:
        fig.savefig(f"{group_name}.lnpz.png")

    return res



#!/usr/bin/env python
def fit_lambda(root='j100025+021706', beams=[], ids=[], newfunc=False, bucket_name='aws-grivam', skip_existing=True, sleep=True, skip_started=True, quasar_fit=False, output_path=None, show_event=False, **kwargs):
    """

    """
    import time
    import os
    import yaml

    import numpy as np
    import boto3
    import json

    if (len(beams) == 0) & (len(ids) == 0):
        beams, files = get_needed_paths(root, bucket_name=bucket_name, skip_existing=skip_existing)
    elif len(ids) > 0:
        # ids key takes precedence over beams
        beams = ['HST/Pipeline/{0}/Extractions/{0}_{1:05d}.beams.fits'.format(root, int(id)) for id in ids]
        if sleep == 1:
            sleep = False
    else:
        if sleep == 1:
            sleep = False

    if len(beams) == 0:
        print('{0}: No beams to fit'.format(root))

        return False
    # Auth to create a Lambda function (credentials are picked up from above .aws/credentials)
    session = boto3.Session()

    # Make sure Lambda is running in the same region as the HST public dataset
    client = session.client('lambda', region_name='us-east-1')

    # Lambda Function names
    func = 'GrizliTestFunction'
    func = 'GrizliLambda-0-12-0-41'
    func = 'GrizliLambda2020'

    print('Lambda function: {0} (newfunc={1})'.format(func, newfunc))

    # Auth to create a Lambda function
    session = boto3.Session()
    client = session.client('lambda', region_name='us-east-1')

    # s3 object = s3://{bucket_name}/{s3_object_path}
    # e.g., 'Pipeline/sparcs0034/Extractions/sparcs0034_00441.beams.fits'
    all_events = []

    NB = len(beams)
    for i, s3_object_path in enumerate(beams):
        print('{0:>5} / {1:>5} : {2}'.format(i+1, NB, s3_object_path))

        if False:
            # Defaults
            skip_started = True  # SKip objects already started
            quasar_fit = False  # Fit with quasar templates

        event = {
              "s3_object_path": s3_object_path,
              "bucket": bucket_name,
              "skip_started": skip_started,
              "quasar_fit": quasar_fit,
            }

        for arg in kwargs:
            event[arg] = kwargs[arg]  # str(arg)

        if output_path is not None:
            if output_path == 'self':
                event['output_path'] = os.path.dirname(s3_object_path)
            else:
                event['output_path'] = output_path

        for k in event:
            if isinstance(event[k], (list, np.ndarray)):
                event[k] = ','.join(['{0}'.format(a) for a in event[k]])
            else:
                pass

                # try:
                #     event[k] = json.dumps(event[k])
                # except:
                #     print('Couldn\'t json.dumps item: ', event[k])

            # if isinstance(event[k], bool):
            #     event[k] = str(event[k])
            #
            # if isinstance(event[k], dict):
            #     event[k] = json.dumps(event[k])

        if show_event == 1:
            print('Event: \n\n', event.__str__().replace('\'', '"').replace(',', ',\n'))
            response = client.invoke(FunctionName=func,
                                     InvocationType='Event', LogType='Tail',
                                     Payload=json.dumps(event))
        if show_event == 2:
            all_events.append(event)
        else:
            # Invoke Lambda function
            response = client.invoke(FunctionName=func,
                                     InvocationType='Event', LogType='Tail',
                                     Payload=json.dumps(event))

    if show_event == 2:
        return all_events

    # Sleep for status check
    beams, full, logs, start = get_needed_paths(root, bucket_name=bucket_name, skip_existing=True, get_lists=True)
    if (sleep) & (len(beams) > len(full)) & (len(beams) > 0):
        sleep_time = 303*np.ceil(len(beams)/950)
        print('{0}: sleep {1}'.format(time.ctime(), sleep_time))

        time.sleep(sleep_time)

        # Status again to check products

        # Wait up to an extra 10 minutes checking if beams finish
        iter = 0
        while (len(beams) > len(full)) & (len(start) > 0) & (iter < 10):
            iter += 1
            time.sleep(61)
            beams, full, logs, start = get_needed_paths(root, bucket_name=bucket_name, skip_existing=True, get_lists=True)


BASE_EVENT = {'bucket': 'grizli-v2', 'skip_started': True, 'quasar_fit': False, 'zr': '0.01,3.2', 'force_args': True}


def generate_events(roots, ids, base=BASE_EVENT, send_to_lambda=False):
    """
    Generate many events with `s3_object_path` based on roots/ids
    """
    events = []
    for root, id in zip(roots, ids):
        event = base.copy()
        event['s3_object_path'] = 'HST/Pipeline/{0}/Extractions/{0}_{1:05d}.beams.fits'.format(root, id)
        events.append(event)

    if send_to_lambda:
        client = get_lambda_client()
        for event in events:
            send_event_lambda(event, verbose=True, client=client, func='GrizliLambda2020')

    return events


def get_lambda_client(region_name='us-east-1'):
    """
    Get boto3 client in same region as HST public dataset
    """
    import boto3
    session = boto3.Session()
    client = session.client('lambda', region_name=region_name)
    return client


def send_event_lambda(event, verbose=True, client=None, func='grizli-redshift-fit'):
    """
    Send a single event to AWS lambda
    
    GrizliLambda2020
    """
    import time
    import os
    import yaml

    import numpy as np
    import boto3
    import json

    if client is None:
        client = get_lambda_client(region_name='us-east-1')

    if ('output_path' in event) & ('s3_object_path' in event):
        if event['output_path'] == 'self':
            event['output_path'] = os.path.dirname(event['s3_object_path'])

    if verbose:
        print('Send event to {0}: {1}'.format(func, event))

    response = client.invoke(FunctionName=func,
                             InvocationType='Event', LogType='Tail',
                             Payload=json.dumps(event))


def get_needed_paths(root, get_string=False, bucket_name='aws-grivam', skip_existing=True, get_lists=False):
    """
    Get the S3 paths of the "beams.fits" files that still need to be fit.
    """
    import boto3
    import time

    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bkt = s3.Bucket(bucket_name)

    files = [obj.key for obj in bkt.objects.filter(Prefix='HST/Pipeline/{0}/Extractions/'.format(root))]

    beams = []
    logs = []
    full = []
    start = []

    for file in files:
        if 'beams.fits' in file:
            beams.append(file)

        if 'log_par' in file:
            logs.append(file)

        if 'start.log' in file:
            start.append(file)

        if 'full.fits' in file:
            full.append(file)

    label = '{0} / {1} / Nbeams: {2}, Nfull: {3}, Nlog: {4}, Nstart: {5}'.format(root, time.ctime(), len(beams), len(full), len(logs), len(start))
    if get_string:
        return label

    print(label)

    if get_lists:
        return beams, full, logs, start

    for i in range(len(beams))[::-1]:
        test = (beams[i].replace('.beams.fits', '.full.fits') in full)
        test |= (beams[i].replace('.beams.fits', '.start.log') in start)
        if test & skip_existing:
            beams.pop(i)

    return beams, files


if __name__ == "__main__":
    import sys
    import yaml

    import numpy as np
    from grizli import utils
    from grizli.pipeline import auto_script
    utils.set_warnings()

    if len(sys.argv) < 2:
        print('Usage: fit_redshift_lambda.py {field}')
        exit

    root = sys.argv[1]

    # bucket_name = 'grizli-grism'
    bucket_name = 'aws-grivam'
    skip_existing = True
    newfunc = False

    kwargs = {'root': root,
              'bucket_name': bucket_name,
              'skip_existing': True,
              'newfunc': False,
              'sleep': True,
              'beams': [],
              'output_path': None,
              'zr': None,
              'show_event': False}

    # Args passed to the lambda event
    kwargs['skip_started'] = True  # SKip objects already started
    kwargs['quasar_fit'] = False  # Fit with quasar templates

    dryrun = False

    if len(sys.argv) > 2:
        for args in sys.argv[2:]:
            keypair = args.strip('--').split('=')

            # Booleans
            if keypair[0] in ['newfunc', 'skip_existing', 'sleep', 'skip_started', 'quasar_fit', 'show_event']:
                if len(keypair) == 1:
                    kwargs[keypair[0]] = True
                else:
                    kwargs[keypair[0]] = keypair[1].lower() in ['true']

            # List of s3 beams.fits paths
            elif keypair[0] == 'beams':
                # Fit a single object
                kwargs['beams'] = keypair[1].split(',')

            # List of ids associated with {root}
            elif keypair[0] == 'ids':
                kwargs['beams'] = ['HST/Pipeline/{0}/Extractions/{0}_{1:05d}.beams.fits'.format(root, int(id)) for id in keypair[1].split(',')]

            # don't run the script
            elif keypair[0] == 'dryrun':
                dryrun = True

            # Everything else
            else:
                if keypair[0] in kwargs:
                    kwargs[keypair[0]] = keypair[1]

    print('Arguments: \n\n', '  '+yaml.dump(kwargs).replace('\n', '\n   '))

    if not dryrun:
        fit_lambda(**kwargs)  # newfunc=newfunc, bucket_name=bucket_name, skip_existing=skip_existing)



"""
Simple tests for grizli
"""

def test_flt():
    """
    Disperse a direct FLT
    """
    import matplotlib as mpl
    mpl.use('Agg')
    
    import matplotlib.pyplot as plt    
    import matplotlib.gridspec
    import numpy as np
    
    import grizli
        
    ### (demo on aligned, background-subtracted FLT images)
    
    #########
    ### Initialize the GrismFLT object
    flt = grizli.model.GrismFLT(flt_file='ibhj34h8q_flt.fits', direct_image='ibhj34h6q_flt.fits', refimage=None, segimage=None, verbose=True, pad=0)
    
    ## Make a catalog/segmetnation image from the direct FLT and make a full
    ## grism model for those detected objects
    flt.photutils_detection(detect_thresh=2, grow_seg=5, gauss_fwhm=2., 
                            compute_beams=['A','B', 'C','D'],
                            verbose=True, save_detection=False, wcs=None)
    
    ## Find object near (x,y) = (495, 749)
    #dr = np.sqrt((flt.catalog['xcentroid']-330)**2+(flt.catalog['ycentroid']-744)**2)
    dr = np.sqrt((flt.catalog['xcentroid']-495)**2+(flt.catalog['ycentroid']-749)**2)
    dr = np.sqrt((flt.catalog['xcentroid']-712)**2+(flt.catalog['ycentroid']-52)**2)
    
    ix = np.argmin(dr)
    id, x0, y0 = flt.catalog['id'][ix], flt.catalog['xcentroid'][ix]+1, flt.catalog['ycentroid'][ix]+1
    
    ## Get basic trace parameters, `conf` is a grizli.grism.aXeConf object, here for G141 & F140W
    # x pixels from the center of the direct image
    dx = np.arange(220)
    # ytrace and wavelength at x=dx
    dy, lam = flt.conf.get_beam_trace(x=x0, y=y0, dx=dx, beam='A')
    
    fig = plt.figure(figsize=[5,1.5])
    #fig = plt.Figure(figsize=[5,1.5])
    ax = fig.add_subplot(111)
    ax.imshow(flt.im_data['SCI'], cmap='gray_r', vmin=-0.05, vmax=0.2, interpolation='Nearest', aspect='auto')
    ax.set_xlim(x0-10, x0+230)
    ax.set_ylim(y0-10, y0+10)
    
    ax.plot(x0+dx-1, y0+dy-1, color='red', linewidth=3, alpha=0.7)
    ## 0.1 micron tick marks along the trace as in the next figure
    xint = np.interp(np.arange(1,1.81,0.1), lam/1.e4, dx)
    yint = np.interp(np.arange(1,1.81,0.1), lam/1.e4, dy)
    ax.scatter(x0+xint-1, y0+yint-1, marker='o', color='red', alpha=0.8)
    ax.set_xlabel(r'$x$ (FLT)'); ax.set_ylabel(r'$y$ (FLT)')
    
    fig.tight_layout(pad=0.1)
    fig.savefig('grizli_demo_0.pdf')
        
    #########
    ### Spectrum cutout
    beam = grizli.model.BeamCutout(id=id, x=x0, y=y0, cutout_dimensions=[18,18], conf=flt.conf, GrismFLT=flt)
    
    # (mask bad pixel)
    beam.cutout_seg[(beam.thumb/beam.photflam > 100) | (beam.thumb < 0)] = 0
    beam.total_flux = np.sum(beam.thumb[beam.cutout_seg == beam.id])
    
    ### Compute the model in the FLT frame for a single object
    model_id = flt.compute_model(id=id, x=x0, y=y0, sh=[80,80], in_place=False).reshape(flt.sh_pad)
    beam.contam = beam.get_cutout(flt.model-model_id)
    
    ## 1D optimal extraction (Horne 1986)
    xspec, yspec, yerr = beam.optimal_extract(beam.cutout_sci, bin=0)
    ## Simple flat flambda continuum model, normalized at F140W
    beam.compute_model(beam.thumb, id=beam.id)
    cmodel = beam.model*1.
    xspecm, yspecm, yerrm = beam.optimal_extract(cmodel, bin=0)
    
    ## Fit simple line + continuum slope
    line_centers, coeffs, chi2, lmask, lmodel, l0, lflux = beam.simple_line_fit(fwhm=5., grid=[1.12e4, 1.65e4, 1, 20])
    xspecl, yspecl, yerrl = beam.optimal_extract(lmodel, bin=0)
    
    ### Make a figure
    fig = plt.figure(figsize=(8,4))
    #fig = plt.Figure(figsize=(8,4))
    
    ## 1D plots
    gsb = matplotlib.gridspec.GridSpec(3,1)  
    
    ax = fig.add_subplot(gsb[-2:,:])
    ax.errorbar(xspec/1.e4, yspec, yerr, linestyle='None', marker='o', markersize=3, color='black', alpha=0.5, label='Data (id=%d)' %(beam.id))
    ax.plot(xspecm/1.e4, yspecm, color='red', linewidth=2, alpha=0.8, label=r'Flat $f_\lambda$ (%s)' %(beam.filter))
    ax.plot(xspecl/1.e4, yspecl, color='orange', linewidth=2, alpha=0.8, label='Cont+line (%.3f, %.2e)' %(l0/1.e4, lflux*1.e-17))
    ax.legend(fontsize=8, loc='lower center', scatterpoints=1)
    
    ax.set_xlabel(r'$\lambda$'); ax.set_ylabel('flux (e-/s)')
    
    ax = fig.add_subplot(gsb[-3,:])
    ax.plot(line_centers/1.e4, chi2/lmask.sum())
    ax.set_xticklabels([])
    ax.set_ylabel(r'$\chi^2/(\nu=%d)$' %(lmask.sum()))
    
    xt = np.arange(1.,1.82,0.1)
    for ax in fig.axes:
        ax.set_xlim(1., 1.8)
        ax.set_xticks(xt)
        
    axt = ax.twiny()
    axt.set_xlim(np.array(ax.get_xlim())*1.e4/6563.-1)
    axt.set_xlabel(r'$z_\mathrm{H\alpha}$')

    ## 2D spectra
    gst = matplotlib.gridspec.GridSpec(3,1)  
    if 'viridis_r' in plt.colormaps():
        cmap = 'viridis_r'
    else:
        cmap = 'cubehelix_r'
    
    ax = fig.add_subplot(gst[0,:])
    ax.imshow(beam.cutout_sci, vmin=-0.05, vmax=0.2, cmap=cmap, interpolation='Nearest', origin='lower', aspect='auto')
    ax.set_ylabel('Observed')
    
    ax = fig.add_subplot(gst[1,:])
    ax.imshow(lmodel+beam.contam, vmin=-0.05, vmax=0.2, cmap=cmap, interpolation='Nearest', origin='lower', aspect='auto')
    ax.set_ylabel('Model')

    ax = fig.add_subplot(gst[2,:])
    ax.imshow(beam.cutout_sci-lmodel-beam.contam, vmin=-0.05, vmax=0.2, cmap=cmap, interpolation='Nearest', origin='lower', aspect='auto')
    ax.set_ylabel('Resid.')
    
    for ax in fig.axes[-3:]:
        ax.set_yticklabels([])
        xi = np.interp(xt, beam.wave/1.e4, np.arange(beam.shg[1]))
        xl = np.interp([1,1.8], beam.wave/1.e4, np.arange(beam.shg[1]))
        ax.set_xlim(xl)
        ax.set_xticks(xi)
        ax.set_xticklabels([])
    
    gsb.tight_layout(fig, pad=0.1,h_pad=0.01, rect=(0,0,0.5,1))
    gst.tight_layout(fig, pad=0.1,h_pad=0.01, rect=(0.5,0.1,1,0.9))
    
    fig.savefig('grizli_demo_1.pdf')
    
    # ## Emission line with calibrated flux
    # lpar = lflux*1.e-17,l0, 3 # line parameters: flux, wave, rms width
    # wave_array = np.arange(lpar[1]-100, lpar[1]+100, lpar[2]/10.)
    # gauss_line = lpar[0]/np.sqrt(2*np.pi*lpar[2]**2)*np.exp(-(wave_array - lpar[1])**2/2/lpar[2]**2)
    # line_model = beam.compute_model(beam.thumb/beam.total_flux, xspec=wave_array, yspec=gauss_line, id=beam.id, in_place=False)
    # xspecl, yspecl, yerrl = beam.optimal_extract(line_model.reshape(beam.shg), bin=0)
    # plt.plot(xspecl, yspecl+yspecm, color='green', linewidth=2, alpha=0.8, linestyle='steps-mid')
    # 
    # ## Check line flux with pysynphot
    # import pysynphot as S
    # g = S.ArraySpectrum(wave_array, gauss_line, fluxunits='flam')
    # g = S.GaussianSource(lflux*1.e-17, l0, lpar[2]*2.35)
    # bp = S.ObsBandpass('wfc3,ir,g141')
    # obs = S.Observation(g, bp) #, force='extrap')
    # print 'Model countrate: %.2f, pySynphot countrate: %.2f' %(line_model.sum(), obs.countrate())
    
    #######################################
    
        
    
    


"""
Tests for JWST imaging and spectra, including 
"""
import os
import glob
import unittest
import numpy as np
import astropy.io.fits as pyfits
import logging
import pytest

from grizli import prep, utils, multifit, GRIZLI_PATH, jwst_utils
from grizli.pipeline import auto_script


def set_crds(path='crds_cache'):
    """
    Set CRDS environment variables if not already set
    """
    if os.getenv('CRDS_SERVER_URL') is None:
        os.environ['CRDS_SERVER_URL'] = 'https://jwst-crds.stsci.edu'
        logging.warn("Set CRDS_SERVER_URL = 'https://jwst-crds.stsci.edu'")
        
    if os.getenv('CRDS_PATH') is None:
        if not path.startswith('/'):
            path = os.path.join(os.getcwd(), path)
        
        os.environ['CRDS_PATH'] = path
            
        logging.warn(f"Set CRDS_PATH = '{path}'")
        if not os.path.exists(path):
            os.mkdir(path)
            os.mkdir(os.path.join(path, 'config'))
            os.mkdir(os.path.join(path, 'config', 'jwst'))
            
set_crds()

class TestJWSTHeaders:
    
    def get_test_data_path(self):
        path = os.path.dirname(utils.__file__)
        path_to_files = os.path.join(path, 'tests/data/jwst-headers/')
        return path_to_files
    
    
    def fetch_test_data(self):
        """
        Fetch test data from MAST
        """
        pass
    
    
    def test_process_headers(self, tmp_path):
        """
        """
        pytest.importorskip('jwst')
                
        file_path = self.get_test_data_path()
        
        files = glob.glob(os.path.join(file_path, '*fits.gz'))
        files.sort()
        
        assert(len(files) == 5)
        
        orig_dir = os.getcwd()
        os.chdir(tmp_path)
        
        # fresh_flt_file
        gz_file = os.path.basename(files[0])
        prep.fresh_flt_file(gz_file, path=file_path)
        
        local_file = gz_file.split('.gz')[0]
        prep.fresh_flt_file(local_file, path=file_path)
        
        # Multiple modes
        for file in files:
            gz_file = os.path.basename(file)
            local_file = gz_file.split('.gz')[0]
            prep.fresh_flt_file(local_file, path=file_path)
            im = pyfits.open(local_file)
            
            assert('SIPRAMAD' in im['SCI'].header)
            assert(im['SCI'].header['SIPRAMAD'] < 1.e-2)
            assert(im['SCI'].header['SIPDEMAD'] < 1.e-2)
            
        # Parse visits
        visits, groups, info = auto_script.parse_visits(field_root='jwst', 
                                                        RAW_PATH=file_path, 
                                                        visit_split_shift=1.2)
        
        assert(os.path.exists('jwst_visits.yaml'))
        
        assert(len(visits) == 5)
        assert(len(groups) == 1)
        
        os.chdir(orig_dir)
    
    
    def test_crds_scale(self, context='jwst_1130.pmap'):
        """
        Test CRDS photometry scaling
        """
        
        pytest.importorskip('jwst')
                
        file_path = self.get_test_data_path()
        
        files = glob.glob(os.path.join(file_path, '*fits.gz'))
        files.sort()
        
        for file in files:
            with pyfits.open(file) as hdul:
                
                mode = {'context': context,
                        'verbose': False,
                        'instrument': hdul[0].header['INSTRUME'],
                        }
                
                for k in ['FILTER','PUPIL','DETECTOR']:
                    if k in hdul[0].header:
                        mode[k.lower()] = hdul[0].header[k]
    
                _ = jwst_utils.get_crds_zeropoint(**mode)
                
                _ = utils.jwst_crds_photom_scale(hdul,
                                                 context=context,
                                                 update=True,
                                                 verbose=True)


class TestJWSTUtils:
    
    def test_filter_info(self):
        """
        Read the info file and get filter data
        """
        import astropy.io.fits as pyfits
        
        bp = jwst_utils.load_jwst_filter_info()
        assert('meta' in bp)
        
        # NIRISS
        header = pyfits.Header()
        header['TELESCOP'] = 'JWST'
        header['INSTRUME'] = 'NIRISS'
        header['PUPIL'] = 'F200W'
        header['FILTER'] = 'CLEAR'
        header['DETECTOR'] = 'NIS'
        
        assert(utils.parse_filter_from_header(header) == 'F200W-CLEAR')
                
        info = jwst_utils.get_jwst_filter_info(header)
        
        assert(info['name'] == 'F200W')
        assert(np.allclose(info['pivot'], 1.992959))
        
        # NIRCam
        header = pyfits.Header()
        header['TELESCOP'] = 'JWST'
        header['INSTRUME'] = 'NIRCAM'
        header['FILTER'] = 'F200W'
        header['PUPIL'] = 'CLEAR'
        header['DETECTOR'] = 'NRCA1'
        
        assert(utils.parse_filter_from_header(header) == 'F200W-CLEAR')
                
        info = jwst_utils.get_jwst_filter_info(header)
        
        assert(info['name'] == 'F200W')
        assert(np.allclose(info['pivot'], 1.988647))
        
        # MIRI
        header = pyfits.Header()
        header['TELESCOP'] = 'JWST'
        header['INSTRUME'] = 'MIRI'
        header['FILTER'] = 'F560W'
        
        assert(utils.parse_filter_from_header(header) == 'F560W')
        
        info = jwst_utils.get_jwst_filter_info(header)
        assert(info['name'] == 'F560W')
        assert(np.allclose(info['pivot'], 5.632612))
    
    
    def test_miri_photom(self):
        
        try:
            import jwst
        except ImportError:
            return None
        
        photmjsr, corr = jwst_utils.get_miri_photmjsr(
                                        file=None,
                                        filter='F770W',
                                        subarray='FULL',
                                        mjd=60153.23,
                                        photom_file='jwst_miri_photom_0201.fits',
                                        verbose=True,
                                    )
        
        assert np.allclose([photmjsr], 0.25890106)
        assert np.allclose([corr], 0.)
        
        # Missing filter
        photmjsr, corr = jwst_utils.get_miri_photmjsr(
                                        file=None,
                                        filter='xF770W',
                                        subarray='FULL',
                                        mjd=60153.23,
                                        photom_file='jwst_miri_photom_0201.fits',
                                        verbose=True,
                                    )
        
        assert np.isnan(photmjsr)


class TestJWSTFittingTools:
    
    def test_config(self):
        """
        Fetch config files if CONF not found
        """
        conf_path = os.path.join(GRIZLI_PATH, 'CONF')
        if not os.path.exists(conf_path): # if CONF dir doesn't already exist, create it
            os.mkdir(conf_path)
        
        if not os.path.exists(os.path.join(conf_path,
                              'GR150C.F115W.conf')):
            print(f'Download config and calib files to {conf_path}')
            #utils.fetch_default_calibs(ACS=False)
            utils.fetch_config_files(get_epsf=True, get_jwst=True)
            files = glob.glob(f'{conf_path}/*')
            print('Files: ', '\n'.join(files))

        assert os.path.exists(os.path.join(conf_path,
                              'GR150C.F115W.conf'))
    
    def test_multibeam(self):
        """
        Can we initialize a multibeam file?
        """
        path = os.path.dirname(utils.__file__)
        print(path)
        beams_file = path + '/tests/data/niriss_jw01324001001_test.beams.fits'
        mb = multifit.MultiBeam(beams_file, group_name='jw01324001001',
                                MW_EBV=-1, fcontam=0.1, sys_err=0.03)
    
        assert(mb.N == 2)
        assert('F115W' in mb.PA)
        
        _ = mb.compute_model()
        
        spec = mb.oned_spectrum()
    
    #def test_parse_visits(self):
    #    """
    #    """
    #    
    #    # These are copies of the WFC3-ERSII demo FLT files but where 
    #    # the data extensions have been set to dtype=uint8 so that they 
    #    # gzip up small
    #    
    #    path = os.path.dirname(utils.__file__)        
    #    if not os.path.exists('niriss-parse-test'):
    #        os.system(f'tar xzvf {path}/tests/data/niriss-parse-test.tar.gz')
    #        
    #    os.chdir('jwst-parse-test/Prep/')
    #    
    #    files = glob.glob('../RAW/*rate.fits')
    #    files.sort()
    #    info = utils.get_flt_info(files)
    #    
    #    assert(len(info) == len(files))
    #    
    #    _visits, _filters = utils.parse_flt_files(info=info, 
    #                                  uniquename=True, get_footprint=True, 
    #                                  use_visit=True, max_dt=0.5, 
    #                                  visit_split_shift=2)
    #    
    #    assert(len(_visits) == 4)
    #    
    #    _groups = utils.parse_grism_associations(_visits)
    #    assert(len(_groups) == 2)
    #    
    #    ### From auto_script
    #    kwargs = auto_script.get_yml_parameters(local_file='my_params.yml', 
    #                                            copy_defaults=False)
    #    
    #    root = 'test'
    #    
    #    for bval in [True, False]:
    #        kwargs['parse_visits_args']['combine_same_pa'] = bval
    #        _ = auto_script.parse_visits(field_root=root,
    #                                     **kwargs['parse_visits_args'])
#
    #        visits, groups, info = _
    #    
    #        assert(hasattr(info, 'colnames'))
    #        assert(len(info) == 16)
    #    
    #        assert(len(visits) == 4)
    #    
    #    assert(os.path.exists(f'{root}_visits.yaml'))
    #    
    #    __ = auto_script.load_visits_yaml(f'{root}_visits.yaml')
    #    _visits, _groups, _info = __
    #    
    #    for v in [visits[0], _visits[0]]:
    #        for k in ['product','files','footprint']:
    #            assert(k in v)
    #    
    #        assert(v['product'].startswith('wfc3-ers'))
    #        assert(len(v['files']) == 4)
    #    
    #    # Grism groups
    #    for g in [groups, _groups]:
    #        assert(len(g) == 2)
    #        
    #        gi = g[0]
    #        for k in ['grism', 'direct']:
    #            assert(k in gi)
    #        
    #        gr = gi['grism'] 
    #        assert(gr['product'] == 'wfc3-ersii-g01-b6o-21-119.0-g102')
    #        assert(len(gr['files']) == 4)
    #        
    #    os.chdir('../../')
    #    if os.path.exists('wfc3-parse-test'):
    #        os.system('rm -rf wfc3-parse-test')
        
class TestGrismConf:
    
    def test_grismconf(self):
        """
        Test initializing ``grizli.grismconf.CRDSGrismConf``
        """
        import grizli.grismconf
        
        pytest.importorskip('jwst')
        
        refs = jwst_utils.crds_reffiles(instrument='NIRCAM',
                                                   filter='F444W',
                                                   pupil='GRISMR',
                                                   module='A',
                                                   date='2023-09-07 00:00:00',
                                                   reftypes=('photom', 'specwcs'), 
                                                   header=None, 
                                                   context='jwst_1123.pmap')
                                                   
        assert('jwst_nircam_specwcs_0190' in refs['specwcs'])
        
        conf = grizli.grismconf.CRDSGrismConf(file=refs['specwcs'], get_photom=False)
        
        conf = grizli.grismconf.CRDSGrismConf(file=refs['specwcs'], get_photom=True)
        



"""
Automated processing of associated exposures
"""


def fetch_from_AWS_bucket(root='j022644-044142', id=1161, product='.beams.fits', bucket_name='aws-grivam', verbose=True, dryrun=False, output_path='./', get_fit_args=False, skip_existing=True):
    """
    Fetch products from the Grizli AWS bucket.

    Boto3 will require that you have set up your AWS credentials in, e.g.,
    ~/.aws/credentials
    """
    import os
    import boto3

    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bkt = s3.Bucket(bucket_name)

    files = [obj.key for obj in bkt.objects.filter(Prefix='Pipeline/{0}/Extractions/{0}_{1:05d}{2}'.format(root, id, product))]

    if get_fit_args:
        files += ['Pipeline/{0}/Extractions/fit_args.npy'.format(root)]

    for file in files:
        local = os.path.join(output_path, os.path.basename(file))

        if verbose:
            print('{0} -> {1}'.format(file, output_path))

        if not dryrun:
            if os.path.exists(local) & skip_existing:
                continue

            bkt.download_file(file, local,
                          ExtraArgs={"RequestPayer": "requester"})
